#---------------------------------------------------------------------------------------
# Singapore, 1st July 2017

# ACADEMIC SOFTWARE LICENSE for PLSDBC
# ************************************

# Copyright:
# Wing-Cheong Wong
# Bioinformatics Institute (BII) A*STAR Singapore

# This software is subjected to license restrictions. The software is
# provided as is. In its present form and without written consent of 
# the copyright holder, the software is provided in accordance with GPL
# (www.gnu.org/licenses/gpl.html). Among other issues, this excludes any
# warranty and indemnity when using this software. For a commercial license,
# please approach the authors (via email to wongwc@bii.a-star.edu.sg).

#---------------------------------------------------------------------------------------

use strict;	# Forces variables to be declared
use configs;
use filefun;
use mathsfun;
use biascorrAlgo;
use biascorrUtil;
use POSIX;

sub alignSeqCountDatawithOutliers {
	my($datafile, $outlierfile, $numexpt, $pointpersegment, $refslope) = @_;

	my $i;
	my $j;
	my @tmp1;
	my @tmp2;
	my $var1;
	my $var2;
	my $var3;
	my $var4;

	# Initialize parameters
	my @numptspersegment=();
	for ($i=0; $i<$numexpt; $i++) { push(@numptspersegment,$pointpersegment); }
	if ($refslope eq "") { $refslope=-1; }

# -----------------------------------------------------------------------------------------------------------------------
#	Basic summary of dataset
# -----------------------------------------------------------------------------------------------------------------------

	my @data; my $header;
	($var1,$header)=filefun::readCountfile($datafile,$numexpt); @data=@$var1;
#	for ($i=0; $i<scalar(@data); $i++) { print $data[$i]."\n"; }

	my @dataOutliers; my $headerdummy;
	($var1,$headerdummy)=filefun::readCountfile($outlierfile,$numexpt); @dataOutliers=@$var1;
#	for ($i=0; $i<scalar(@dataOutliers); $i++) { print $dataOutliers[$i]."\n"; }

	my $outfile="";
	my $idx=-1; my $refidx=-1;
	my @mincountPerExpt; my @maxcountPerExpt; my @totalreadsPerExpt; my @maxcountPerExptIdx;
	my @intervallist=(); my $maxinterval=-1;
	my @regressionlist=();

	print "1.Basic summary of dataset:\n";

	# Create log rank log count datapoints of original data
	my @origParetoCDFpoints; my @origZipfCDFpoints;
	($var1,$var2)=biascorrUtil::createPowerLawPlotData(\@data,$numexpt);
	@origParetoCDFpoints=@$var1; @origZipfCDFpoints=@$var2;

	# Compute the unique count values of the logrank-logcount plot for each experiment
	my @uniqueptsperexpt=();
	for ($i=0; $i<$numexpt; $i++) { push(@uniqueptsperexpt,0); }
	for ($i=0; $i<scalar(@origParetoCDFpoints); $i++) {
		@tmp2=split(/\t/,$origParetoCDFpoints[$i]);
		for ($j=1; $j<scalar(@tmp2); $j++) { if ($tmp2[$j]>0) { $uniqueptsperexpt[$j-1]=$uniqueptsperexpt[$j-1]+1; } }
	}
	print "No. of unique points in log-log plot:\n"; for ($i=0; $i<$numexpt; $i++) { print "Expt[".$i."]\t".$uniqueptsperexpt[$i]."\n"; } print "\n";

	# Find the smallest, largest read count and total reads for each experiment
	print "Maximum/Minimum count by experiment[Pre-scaling]:\n";
	($var1,$var2,$var3,$var4)=biascorrUtil::createExptDataSummary(\@data,$numexpt,1);
	@mincountPerExpt=@$var1; @maxcountPerExpt=@$var2; @totalreadsPerExpt=@$var3; @maxcountPerExptIdx=@$var4;

	for ($i=0,$refidx=-1; $i<scalar(@maxcountPerExpt); $i++) {
		if ($refidx==-1) { $var1=$maxcountPerExpt[$i]; $refidx=$i; }
		else { if ($var1<$maxcountPerExpt[$i]) { $var1=$maxcountPerExpt[$i]; $refidx=$i; } }
	}
	print "Expt index with maximum count:".$refidx."\n\n";

	# Default parameters
	my $logbase=log($configs::BASEOFLOG);
	if (scalar(@numptspersegment)==0) {
		for ($i=0; $i<$numexpt; $i++) {
			$var1 = floor(($configs::MAXPTSPERSEGMENT/$uniqueptsperexpt[$refidx])*$uniqueptsperexpt[$i]);
			push(@numptspersegment,$var1);
		}
	}
	print "No.of points per segment:\n"; for ($i=0; $i<scalar(@numptspersegment); $i++) { print "Expt[".$i."]\t".$numptspersegment[$i]."\n"; } print "\n";

# -----------------------------------------------------------------------------------------------------------------------
#	Scaling correction of dataset using Zipf's law
# -----------------------------------------------------------------------------------------------------------------------

	# Prior to shifting, segment the logCount-logRank plot into N pieces to obtain the intervals for each experiment. Process for each experiment
	print "2.Perform Scaling [Zipf's law] of dataset:\n";
	@intervallist=(); $maxinterval=-1;
	($var1,$var2) = biascorrUtil::computeIntervalsByPtsperSeg(\@data, $numexpt, $logbase, \@numptspersegment);
	@intervallist=@$var1; $maxinterval=$var2;

	my @intervalsperExpt=();
	for ($i=0; $i<$numexpt; $i++) {
		# Create a sublist of interval for per experiment's log-log plot. Process each segmented log-log plot a time
		for ($j=0,@intervalsperExpt=(); $j<scalar(@intervallist); $j++) {
			@tmp1 = split(/\;/, $intervallist[$j]);	# intervallist ~ [expt;segmentid;startvalue;stopvalue]
			if ($tmp1[0]==$i) { push(@intervalsperExpt, $intervallist[$j]); }
		}

		if (scalar(@intervalsperExpt)>0) {	# For each segment of the same experiment
			# Obtain regression parameters for each interval
			($var1,$var2)=biascorrAlgo::computeRegressionParameters(\@data, $numexpt, \@intervalsperExpt, $logbase);
			@regressionlist=@$var1; @intervalsperExpt=@$var2;

			# Perform scaling of datapoints per experiment
			if (scalar(@regressionlist)>0) {
				print "Regression list for expt[".$i."] per segment[numexpt;segmentID;range;slope;intercept;Rsq;size]:\n";
				for ($j=0; $j<scalar(@regressionlist); $j++) { print $regressionlist[$j]."\n"; }
				$var1=biascorrAlgo::scaleDatatoZipfLaw(\@data,\@regressionlist,$maxcountPerExpt[$i],$refslope,$configs::BASEOFLOG); @data=@$var1;
				$var1=biascorrAlgo::scaleDatatoZipfLaw(\@dataOutliers,\@regressionlist,$maxcountPerExpt[$i],$refslope,$configs::BASEOFLOG); @dataOutliers=@$var1;
			}
		}
	}
	print "\n";


	# Find the smallest, largest read count and total reads for each experiment
	print "Maximum/Minimum count by experiment[Post-scaling]:\n";
	($var1,$var2,$var3,$var4)=biascorrUtil::createExptDataSummary(\@data,$numexpt,1);
	@mincountPerExpt=@$var1; @maxcountPerExpt=@$var2; @totalreadsPerExpt=@$var3; @maxcountPerExptIdx=@$var4;

	# Write the scaled data to file
	@tmp1 = split(/\//,$datafile);
	$outfile = "scale".$tmp1[scalar(@tmp1)-1];
	print "Write to scaled data to output file: ".$outfile."\n\n";
	filefun::writeCountfile($header,\@data,$outfile);

	@tmp1 = split(/\//,$outlierfile);
	$outfile = "scale".$tmp1[scalar(@tmp1)-1];
	print "Write to scaled data to output file: ".$outfile."\n\n";
	filefun::writeCountfile($header,\@dataOutliers,$outfile);

	@tmp1=(); push(@tmp1,@dataOutliers); push(@tmp1,@data); @data=@tmp1;

	# Create log rank log count datapoints of normalized data
	my @scaleParetoCDFpoints; my @scaleZipfCDFpoints;
	($var1,$var2)=biascorrUtil::createPowerLawPlotData(\@data,$numexpt);
	@scaleParetoCDFpoints=@$var1; @scaleZipfCDFpoints=@$var2;

	# Perform regression test per segment to check if the slopes and intercepts are close to the presets
	print "Check regression parameters per segment of each experiment[Post scaling]:\n";
	@intervallist=(); $maxinterval=-1;
	($var1,$var2) = biascorrUtil::computeIntervalsByPtsperSeg(\@data, $numexpt, $logbase, \@numptspersegment);
	@intervallist=@$var1; $maxinterval=$var2;

	for ($i=0,@tmp2=(); $i<$numexpt; $i++) {
		# Extract the list of segmented intervals for each experiment
		for ($j=0,@intervalsperExpt=(); $j<scalar(@intervallist); $j++) {
			@tmp1 = split(/\;/, $intervallist[$j]);	# intervallist ~ [expt;segmentid;startvalue;stopvalue]
			if ($tmp1[0]==$i) { push(@intervalsperExpt, $intervallist[$j]); }
		}

		if (scalar(@intervalsperExpt)>0) {	# For each segment of the same experiment
			# Obtain regression parameters for each interval
			($var1,$var2)=biascorrAlgo::computeRegressionParameters(\@data, $numexpt, \@intervalsperExpt, $logbase);
			@regressionlist=@$var1; @intervalsperExpt=@$var2;
			push(@tmp2, @regressionlist);	# Store each experiment per iteration
		}
	}
	@regressionlist=(); push(@regressionlist,@tmp2);

	print "Regression list per experiment/segment[numexpt;segmentID;range;slope;intercept;Rsq;size]:\n";
	for ($i=0; $i<scalar(@regressionlist); $i++) { print $regressionlist[$i]."\n"; } print "\n";

# -----------------------------------------------------------------------------------------------------------------------

	# Find the best set of segments per experiment to fit Zipf's law (slope=-1)
	print "3.Perform Shifting [to align] of the dataset:\n";
	$var1=biascorrAlgo::computeOptimumShiftfactor(\@regressionlist, \@data, $logbase);
	@regressionlist=@$var1;
	print "Regression list per experiment[numexpt;segmentID;range;slope;intercept;Rsq;size]:\n";
	for ($i=0; $i<scalar(@regressionlist); $i++) { print $regressionlist[$i]."\n"; } print "\n";

	my @maxshiftfactor=();
	for ($i=0; $i<scalar(@maxcountPerExpt); $i++) {
		push(@maxshiftfactor,$maxcountPerExpt[$refidx]/$maxcountPerExpt[$i]);
	}

	# Compute for the shift between log-log plots
	my @intercept=(); my $interceptRef;
	for ($i=0; $i<scalar(@regressionlist); $i++) {
		@tmp1 = split(/\;/, $regressionlist[$i]);	# numexpt;intervalID;range;slope;intercept;Rsq;nsize
		push(@intercept, $tmp1[4]);
	}

	my @shiftfactor=();
	for ($i=0; $i<scalar(@intercept); $i++) { $shiftfactor[$i]=$configs::BASEOFLOG**($intercept[$refidx]-$intercept[$i]); }
	print "Calculated shiftfactors:\n"; for ($i=0; $i<scalar(@shiftfactor); $i++) { print "Expt:".$i."\t".$shiftfactor[$i]."\t".$maxshiftfactor[$i]."\n"; } print "\n";

	# Perform systematic shifts along the log count axis (X-axis) with reference to the logCount-logRank plot with the largest reads
	for ($i=0; $i<$numexpt; $i++) {
		$idx = $i+1;
		for ($j=0; $j<scalar(@data); $j++) {
			@tmp1=split(/\t/, $data[$j]);
			if ($tmp1[$idx]>1) { $tmp1[$idx]=floor($tmp1[$idx]*$shiftfactor[$i]); $data[$j]=join("\t",@tmp1); }
		}
	}
	# Find the smallest, largest read count and total reads for each experiment
	print "Maximum/Minimum count by experiment[Post shifting]:\n";
	($var1,$var2,$var3,$var4)=biascorrUtil::createExptDataSummary(\@data,$numexpt,1);
	@mincountPerExpt=@$var1; @maxcountPerExpt=@$var2; @totalreadsPerExpt=@$var3; @maxcountPerExptIdx=@$var4;

	# Shift minimum count in dataset towards zero
	@tmp1 = sort {$a <=> $b} @mincountPerExpt;
	my $offset=$tmp1[0];
	print "Lowest count:\t".$offset."\n\n";

	# Create log rank log count datapoints of normalized data
	my @normParetoCDFpoints; my @normZipfCDFpoints;
	($var1,$var2)=biascorrUtil::createPowerLawPlotData(\@data,$numexpt);
	@normParetoCDFpoints=@$var1; @normZipfCDFpoints=@$var2;

	print "Check regression parameters per segment of each experiment[Post shifting]:\n";
	@intervallist=(); $maxinterval=-1;
	($var1,$var2) = biascorrUtil::computeIntervalsByPtsperSeg(\@data, $numexpt, $logbase, \@numptspersegment);
	@intervallist=@$var1; $maxinterval=$var2;
#	for ($i=0; $i<scalar(@intervallist); $i++) { print $intervallist[$i]."\n"; } print "\n";

	for ($i=0,@tmp2=(); $i<$numexpt; $i++) {
		# Extract the list of segmented intervals for each experiment
		for ($j=0,@intervalsperExpt=(); $j<scalar(@intervallist); $j++) {
			@tmp1 = split(/\;/, $intervallist[$j]);	# intervallist ~ [expt;segmentid;startvalue;stopvalue]
			if ($tmp1[0]==$i) { push(@intervalsperExpt, $intervallist[$j]); }
		}

		if (scalar(@intervalsperExpt)>0) {	# For each segment of the same experiment
			# Obtain regression parameters for each interval
			($var1,$var2)=biascorrAlgo::computeRegressionParameters(\@data, $numexpt, \@intervalsperExpt, $logbase);
			@regressionlist=@$var1; @intervalsperExpt=@$var2;
			push(@tmp2, @regressionlist);	# Store each experiment per iteration
		}
	}
	@regressionlist=(); push(@regressionlist,@tmp2);

	print "Regression list per experiment/segment[numexpt;segmentID;range;slope;intercept;Rsq;size]:\n";
	for ($i=0; $i<scalar(@regressionlist); $i++) { print $regressionlist[$i]."\n"; } print "\n";

# -----------------------------------------------------------------------------------------------------------------------

	# Perform multiple testing of slopes (See Jerrold Zar, Biostatistical analysis, Chapter 18.4, pp370-372)
	print "4.Test significance among each set of regression lines for each interval:\n";
	my $significantflag; my $Fstatistics; my $Fcritical; my $numeratorDF; my $denominatorDF;
	my @commonIntervals=(); my @untested=(); my $status="";
	my @Xlist=(); my @Ylist=();
	for ($i=0,@untested=(); $i<$maxinterval; $i++) {
		($var1, $var2, $var3) = biascorrUtil::getIntervaldataAcrossExpt(\@data, $numexpt, \@intervallist, $i, $logbase);
		@Xlist=@$var1; @Ylist=@$var2; @commonIntervals=@$var3;

		for ($j=0; $j<scalar(@commonIntervals); $j++) { print $commonIntervals[$j]."\n"; }

		($significantflag, $Fstatistics, $Fcritical, $numeratorDF, $denominatorDF) = mathsfun::testmultipleslopes(\@Xlist, \@Ylist, './statsTable/Fstatsp001.txt');
		if ($significantflag ne "" && $Fstatistics ne "" && $Fcritical ne "" && $numeratorDF ne "" && $denominatorDF ne "") {
			print $Fstatistics." vs ".$Fcritical."[".$numeratorDF.",".$denominatorDF."]\t";
			if ($significantflag == 1) { print "Significant\n"; }
			elsif ($significantflag == 0) { print "Insignificant\n"; }
			else { print "Failed\n"; }
		}
		else {
			push(@untested,@commonIntervals);
			$status = "Untestable";
			print "Untestable\n";
		}
		print "\n";
	}
	if (scalar(@untested)==0) { push(@untested,@commonIntervals); }

# -----------------------------------------------------------------------------------------------------------------------

	# Write the final shifted and scaled data to file
	@tmp1 = split(/\//,$datafile);
	$outfile = "norm".$tmp1[scalar(@tmp1)-1];

	print "5.Write to scaled data to output file: ".$outfile."\n\n";
	filefun::writeCountfile($header,\@data,$outfile);

	# Write the Zipf CDF of original data to file
	@tmp2=split(/\t/,$header); my $header1="Rank";
	for ($i=1; $i<scalar(@tmp2); $i++) { $header1=$header1."\t".$tmp2[$i]."-count"; }

	@tmp1 = split(/\//,$datafile);
	$outfile = "zipfCDFOrig".$tmp1[scalar(@tmp1)-1];
	print "6.Write to zipfCDF data to output file: ".$outfile."\n\n";
	filefun::writefile($header1,\@origZipfCDFpoints,$outfile);

	# Write the Pareto CDF of original/scale/norm data to file
	@tmp2=split(/\t/,$header); my $header2=$tmp2[1]."-count\t".$tmp2[1]."-CDF";
	for ($i=2; $i<scalar(@tmp2); $i++) { $header2=$header2."\t".$tmp2[$i]."-count\t".$tmp2[$i]."-CDF"; }

	@tmp1 = split(/\//,$datafile);
	$outfile = "paretoCDFOrig".$tmp1[scalar(@tmp1)-1];
	print "7.Write to paretoCDF data to output file: ".$outfile."\n\n";
	filefun::writefile($header2,\@origParetoCDFpoints,$outfile);

	$outfile = "paretoCDFScale".$tmp1[scalar(@tmp1)-1];
	print "8.Write to paretoCDF data to output file: ".$outfile."\n\n";
	filefun::writefile($header2,\@scaleParetoCDFpoints,$outfile);

	$outfile = "paretoCDFNorm".$tmp1[scalar(@tmp1)-1];
	print "9.Write to paretoCDF data to output file: ".$outfile."\n\n";
	filefun::writefile($header2,\@normParetoCDFpoints,$outfile);
}

1;