#---------------------------------------------------------------------------------------
# Singapore, 1st July 2017

# ACADEMIC SOFTWARE LICENSE for PLSDBC
# ************************************

# Copyright:
# Wing-Cheong Wong
# Bioinformatics Institute (BII) A*STAR Singapore

# This software is subjected to license restrictions. The software is
# provided as is. In its present form and without written consent of 
# the copyright holder, the software is provided in accordance with GPL
# (www.gnu.org/licenses/gpl.html). Among other issues, this excludes any
# warranty and indemnity when using this software. For a commercial license,
# please approach the authors (via email to wongwc@bii.a-star.edu.sg).

#---------------------------------------------------------------------------------------

#!/usr/bin/perl
package biascorrUtil;
use strict;	# Forces variables to be declared
use warnings;
use POSIX;

sub filterZeroCountData {
	my($dataRef) = @_;
	my @data=@$dataRef;

	my $i; my $j;
	my @tmp1; my $found;

	my @subdata=();
	for ($i=0; $i<scalar(@data); $i++) {
		@tmp1=split(/\t/, $data[$i]);
		for ($j=1,$found=0; $j<scalar(@tmp1); $j++) {	# Offset by 1 due to transcript name
			if ($tmp1[$j]==1) { $found=1; last; }		# Mark the transcript row that has pseudocount of 1 (equivalent to zero)
		}
		if ($found==0) { push(@subdata,$data[$i]); }
	}
	return (\@subdata);
}

sub selectDataCol {
	my ($dataRef,$column) = @_;
	my @data=@$dataRef;

	my $i; my @tmp1;
	my @array=();
	for ($i=0; $i<scalar(@data); $i++) {
		@tmp1=split(/\t/,$data[$i]);
		push(@array,$tmp1[$column]);
	}
#	print $column."\t".scalar(@data)."\t".scalar(@array)."\n";
	return \@array;
}

sub createPowerLawPlotData {
	my ($dataRef,$numexpt) = @_;
	my @data=@$dataRef;	# Assumes data comes with header

	my @logcdfcountpoints=(); my @logrankcountpoints=();

	my $i;
	my $j;
	my $k;
	my $count;
	my @tmp1;
	my @tmp2;
	my @tmp3;
	my $var1;
	my $var2;
	my $max; my $idx;
	my $rank;
	my $transcriptName;

	# Create a rank of the count values in the all experiments
	my %hashtablePerExpt=();
	my @transcriptcountPerExpt=(); my @transcriptcountAllExpt=();	# Zipf's: Each count is a transcript value (might be repeated)
	my @uniquecountPerExpt=(); my @uniquecountAllExpt=();			# Pareto: Each count is a unique value
	my @cummulativecountPerExpt=(); my @cummulativecountAllExpt=();

	for ($i=0; $i<$numexpt; $i++) {
		# Generate data assignment for Pareto and Zipf's log-log plots
		for ($j=0,%hashtablePerExpt=(),@transcriptcountPerExpt=(); $j<scalar(@data); $j++) {
			@tmp1=split(/\t/,$data[$j]);
			$transcriptName=$tmp1[0];
			$count=$tmp1[$i+1];	# offset due to gene name in first column
			if ($count > 1) {
				if (exists($hashtablePerExpt{$count})) { $hashtablePerExpt{$count}=$hashtablePerExpt{$count}+1; } else { $hashtablePerExpt{$count}=1; }
				push(@transcriptcountPerExpt,$count);
			}
		}
		# Pareto distribution: Create a sorted count (largest to smallest) per experiment and compute the cummulative count
		@uniquecountPerExpt=(); @uniquecountPerExpt=sort {$b<=>$a} keys %hashtablePerExpt;
		push(@uniquecountAllExpt,join(";",@uniquecountPerExpt));
		for ($j=0,@cummulativecountPerExpt=(); $j<scalar(@uniquecountPerExpt); $j++) {
			for ($k=0,$var1=0; $k<=$j; $k++) { $var1=$var1+$hashtablePerExpt{$uniquecountPerExpt[$k]}; }
			push(@cummulativecountPerExpt,$var1);
		}
		push(@cummulativecountAllExpt,join(";",@cummulativecountPerExpt));

		# Zipf's distribution: Create a sorted count (largest to smallest) per experiment and compute the cummulative count
		@transcriptcountPerExpt=sort {$b<=>$a} @transcriptcountPerExpt;
		push(@transcriptcountAllExpt,join(";",@transcriptcountPerExpt));
	}

	# Find the longest log-log dataset
	@tmp1=split(/\;/, $uniquecountAllExpt[0]); $max=scalar(@tmp1);
	for ($i=1; $i<scalar(@uniquecountAllExpt); $i++) {
		@tmp1=split(/\;/, $uniquecountAllExpt[$i]); $var1=scalar(@tmp1);
		if ($max<$var1) { $max=$var1; $idx=$i; }
	}
	# Convert to N-by-M matrix output
	for ($i=0,@logcdfcountpoints=(); $i<$max; $i++) {
		for ($j=0,@tmp2=(); $j<scalar(@uniquecountAllExpt); $j++) {
			@tmp1=split(/\;/,$uniquecountAllExpt[$j]);
			if ($i<scalar(@tmp1)) { $var1=$tmp1[$i]; } else { $var1=0; }
			push(@tmp2,$var1);

			@tmp1=split(/\;/,$cummulativecountAllExpt[$j]);
			if ($i<scalar(@tmp1)) { $var1=$tmp1[$i]; } else { $var1=0; }
			push(@tmp2,$var1);
		}
		push(@logcdfcountpoints,$i."\t".join("\t",@tmp2));
	}
#	for ($i=0; $i<scalar(@logcdfcountpoints); $i++) { print $logcdfcountpoints[$i]."\n" }

	# Find the longest log-log dataset
	@tmp1=split(/\;/, $transcriptcountAllExpt[0]); $max=scalar(@tmp1);
	for ($i=1; $i<scalar(@transcriptcountAllExpt); $i++) {
		@tmp1=split(/\;/, $transcriptcountAllExpt[$i]); $var1=scalar(@tmp1);
		if ($max<$var1) { $max=$var1; $idx=$i; }
	}

	# Convert to N-by-M matrix output
	for ($i=0,@logrankcountpoints=(); $i<$max; $i++) {
		$rank=$i+1; @tmp2=(); push(@tmp2,$rank);	# Initialize and push rank for each row across experiments
		for ($j=0; $j<scalar(@transcriptcountAllExpt); $j++) {
			@tmp1=split(/\;/,$transcriptcountAllExpt[$j]);
			if ($i<scalar(@tmp1)) { $var1=$tmp1[$i]; } else { $var1=0; }
			push(@tmp2,$var1);
		}
		push(@logrankcountpoints,join("\t",@tmp2));
	}
#	for ($i=0; $i<scalar(@logrankcountpoints); $i++) { print $logrankcountpoints[$i]."\n" }

	return (\@logcdfcountpoints,\@logrankcountpoints);
}

sub createExptDataSummary {
	my ($dataRef,$numexpt,$flag) = @_;
	my @data=@$dataRef;	# Assumes data comes with header

	my $i;
	my $j;
	my @tmp1;
	my $idx;
	my $count;

	my $mincount; my $maxcount; my $maxidx; my $totalreads;
	my @mincountPerExpt; my @maxcountPerExpt;
	my @maxcountPerExptIdx; my @totalreadsPerExpt;

	# Find the largest read count and total reads for each experiment
	for ($i=0,@mincountPerExpt=(),@maxcountPerExpt=(),@maxcountPerExptIdx=(),@totalreadsPerExpt=(); $i<$numexpt; $i++) {
		$idx = $i+1;
		for ($j=0,$totalreads=0,$mincount=-1,$maxcount=-1,$maxidx=""; $j<scalar(@data); $j++) {
			@tmp1=split(/\t/,$data[$j]);
			$count=$tmp1[$idx];
			if ($count>1) { $totalreads=$totalreads+$count; }	# Compute total reads in experiment, excluding pseudocount of 1
			if ($count>1) {if ($mincount==-1) { $mincount=$count; } else { if ($mincount>$count) { $mincount=$count; }}}			# Find minimum count in experiment
			if ($maxcount==-1) { $maxcount=$count; $maxidx=$j; } else { if ($maxcount<$count) { $maxcount=$count; $maxidx=$j; } }	# Find maximum count in experiment
		}
		push(@mincountPerExpt, $mincount);
		push(@maxcountPerExpt, $maxcount);
		push(@maxcountPerExptIdx,$maxidx);
		push(@totalreadsPerExpt, $totalreads);
	}

	if ($flag==1) {
		for ($i=0; $i<scalar(@maxcountPerExpt); $i++) {
			print $i."\t".$totalreadsPerExpt[$i]."[Total Reads(per million)]\t".$maxcountPerExpt[$i]."[row:".$maxcountPerExptIdx[$i]."],".$mincountPerExpt[$i]."=".($maxcountPerExpt[$i]-$mincountPerExpt[$i]+1)."\n";
		} print "\n";
	}

	return (\@mincountPerExpt,\@maxcountPerExpt,\@totalreadsPerExpt,\@maxcountPerExptIdx);
}

sub getIntervaldatapts {
	my($dataRef, $numexpt, $intervallistRef, $logbase) = @_;
	my @data=@$dataRef;
	my @intervallist=@$intervallistRef;

	my $i;
	my $j;
	my $k;
	my @tmp1;
	my @tmp2;
	my $var1;
	my $var2;
	my $var3;
	my $exptidx;
	my @keys=(); my @sortcount=(); my %hashtable1=(); my @cummulatedCount=();
	my $count; my $rank;
	my @logCount=(); my @logRank=();

	my @Xlist=(); my @Ylist=();	# Return results

	for ($i=0; $i<scalar(@intervallist); $i++) {
		@tmp1 = split(/\;/, $intervallist[$i]);		# numexpt;intervalID;range
		$exptidx = $tmp1[0]+1;						# Get experiment column wrt data array
		@tmp2 = split(/\,/,$tmp1[2]); $var1=$tmp2[0]; $var2=$tmp2[1];	# Get the interval range

		%hashtable1=(); @keys=();
		for ($j=0; $j<scalar(@data); $j++) {
			@tmp1 = split(/\t/, $data[$j]);	# 'Per transcript count' as the key of hashtable, 'No. of transcripts with this count' as the value of hashtable
			# Convert the transcript count to logarithmic scale
			$count = log($tmp1[$exptidx])/$logbase;

			# Capture all datapoints of this experiment
			if (exists($hashtable1{$count})) { $hashtable1{$count}=$hashtable1{$count}+1; }
			else { $hashtable1{$count}=1; }
		}

		# Sort the unique transcript count in descending order (rank in increasing order) and compute the cummulative count of transcript with count>=x
		@sortcount = sort { $b<=>$a } keys %hashtable1;
		for ($j=0,@cummulatedCount=(); $j<scalar(@sortcount); $j++) {
			for ($k=0,$var3=0; $k<=$j; $k++) { $var3=$var3+$hashtable1{$sortcount[$k]}; }
			push(@cummulatedCount,$var3);
		}

		# Get the datapoints within this interval range for this experiment
		for ($j=0,@logCount=(),@logRank=(); $j<scalar(@sortcount); $j++) {
			if ($sortcount[$j]<=$var2 && $sortcount[$j]>$var1) {
#				$rank = $j+1;
				$rank=$cummulatedCount[$j];		# Based on Pareto's cdf
				push(@logCount, $sortcount[$j]);
				push(@logRank, log($rank)/$logbase);
			}
		}

		# For each experiment, store the datapoints within this interval into arrays of X and Y
		push(@Xlist, join(';', @logCount));
		push(@Ylist, join(';', @logRank));
#		print $intervallist[$i]."\t".scalar(@logCount)."\t".scalar(@logRank)."\n";
	}
	return (\@Xlist, \@Ylist);
}

sub computeIntervalsByPtsperSeg {
	my($dataRef, $numexpt, $logbase, $pointspersegmentRef) = @_;
	my @data=@$dataRef;
	my @pointspersegment=@$pointspersegmentRef;

	my $i;
	my $j;
	my $k;
	my $count=0;
	my @tmp1;
	my @tmp2;
	my $var1;
	my $var2;
	my $var3;
	my $idx=-1;
	my $rank; my @sortcount=(); my %hashtable1=(); my @cummulatedCount=();
	my $numintervals; my $maxlogcount; my $minlogcount;
	my @logCount=(); my @logRank=();
	my @intervallist=(); my $maxinterval=-1;	# Return results

	# For each experiment, segment the logCount-logRank plot into N intervals
	@intervallist=(); $maxinterval=-1;
	for ($i=0; $i<$numexpt; $i++) {
		$idx = $i+1; %hashtable1=();

		# 'Per transcript count' as the key of hashtable, 'Name of transcripts with this count' as the value of hashtable
		for ($j=0; $j<scalar(@data); $j++) {
			@tmp1 = split(/\t/, $data[$j]);
			$count = $tmp1[$idx];

			# Collect the reads (exclude pseudo-read value of 1)
			if ($count>1) {
				if (exists($hashtable1{$count})) { $hashtable1{$count}=$hashtable1{$count}+1; }
				else { $hashtable1{$count}=1; }
			}
		}

		# Sort the unique transcript count in descending order (rank in increasing order) and compute the cummulative count of transcript with count>=x
		@sortcount = sort { $b<=>$a } keys %hashtable1;
		for ($j=0,@cummulatedCount=(); $j<scalar(@sortcount); $j++) {
			for ($k=0,$var1=0; $k<=$j; $k++) { $var1=$var1+$hashtable1{$sortcount[$k]}; }
			push(@cummulatedCount,$var1);
		}

#		print "Index/t/Unique-count\tNo.ofcount\tCummulative-count\tExpt[".$i."]\n";
#		for ($j=0; $j<scalar(@sortcount); $j++) {
#			print $j."\t".$sortcount[$j]."\t".$hashtable1{$sortcount[$j]}."\t".$cummulatedCount[$j]."\n";
#		}

		# Create the logCount and logRank arrays
		for ($j=0,@logCount=(),@logRank=(); $j<scalar(@sortcount); $j++) {
#			$rank = $j+1;
			$rank=$cummulatedCount[$j];		# Based on Pareto's cdf
			push(@logCount, log($sortcount[$j])/$logbase);
			push(@logRank, log($rank)/$logbase);
		}
		$numintervals = floor(scalar(@logCount)/$pointspersegment[$i])+1;

		# Get the minimum and maximum log count
		@tmp1 = sort {$a<=>$b} @logCount;
		$minlogcount=$tmp1[0]; $maxlogcount=$tmp1[scalar(@tmp1)-1];

		for ($j=0,$var1=0,$var2=0; $j<$numintervals; $j++) {
			$var1=$j*$pointspersegment[$i];
			$var2=($j+1)*$pointspersegment[$i];
			if ($var2>scalar(@logCount)-1) { $var2=scalar(@logCount)-1; }	# limit the max index
			if ($var2>$var1) {
				push(@intervallist, $i.";".$j.";".$logCount[$var2].",".$logCount[$var1]);
			}
		}
		if ($maxinterval < $j) { $maxinterval = $j; }
	}

#	for ($i=0; $i<scalar(@intervallist); $i++) { print $intervallist[$i]."\n"; }
#	print "Maximum interval: ".$maxinterval."\n\n";

	return (\@intervallist, $maxinterval);
}

sub getIntervaldataAcrossExpt {
	my($dataRef, $numexpt, $intervallistRef, $intervalID, $logbase) = @_;
	my @data=@$dataRef;
	my @intervallist=@$intervallistRef;

	my $j;
	my $k;
	my $found;
	my $count;
	my @tmp1;
	my @tmp2;
	my $var1;
	my $var2;
	my $idx;
	my $exptidx;
	my $rank; my @keys=(); my @sortkeys=(); my %hashtable1=();
	my @logCount=(); my @logRank=();

	my @Xlist=(); my @Ylist=(); my @commonIntervals=();	# Return results

#	print "For test interval-".$intervalID."'s ranges:\n";
	for ($j=0,@Xlist=(),@Ylist=(); $j<$numexpt; $j++) {
		# Find the same interval across different experiments
		for ($k=0,$found=0,$idx=-1; $k<scalar(@intervallist); $k++) {
			@tmp1 = split(/\;/, $intervallist[$k]);	# e.g. 0;0;5.603,6.603;-0.97;6.51;0.96;11 i.e. numexpt;intervalID;range;slope;intercept;Rsq;n
			if ($tmp1[0]==$j && $tmp1[1]==$intervalID) {	# Same interval across different experiments
				$found=1; $idx=$k;
				last;
			}
		}

		# For the same interval across different experiment
		if ($found==1) {
			# Save the selected interval for the experiment
			push(@commonIntervals, $intervallist[$idx]);

#			print $intervallist[$idx]."\n";
			@tmp2 = split(/\;/, $intervallist[$idx]);	# e.g. 0;0;5.603,6.603 i.e. numexpt;intervalID;range
			$exptidx = $tmp2[0]+1;	# Get experiment column wrt data array
			@tmp1 = split(/\,/,$tmp2[2]); $var1=$tmp1[0]; $var2=$tmp1[1]; # Get the interval range
			%hashtable1=(); @keys=();
			for ($k=0; $k<scalar(@data); $k++) {
				@tmp1 = split(/\t/, $data[$k]);	# 'Per transcript count' as the key of hashtable, 'Name of transcripts with this count' as the value of hashtable
				# Convert the transcript count to logarithmic scale
				$count = log($tmp1[$exptidx])/$logbase;

				# Capture only those datapoints relevant for this experiment
				if (exists($hashtable1{$count})) { $hashtable1{$count} = $hashtable1{$count}.";".$tmp1[0]; }
				else { $hashtable1{$count} = $tmp1[0]; }
			}

			# Sort the unique transcript in descending order (rank in increasing order)
			@keys = keys %hashtable1;
			@sortkeys = sort { $b<=>$a } @keys;

			# Get the datapoints within this interval range for this experiment
			for ($k=0,@logCount=(),@logRank=(); $k<scalar(@sortkeys); $k++) {
				if ($sortkeys[$k]<=$var2 && $sortkeys[$k]>$var1) {
					$rank = $k+1;
					push(@logCount, $sortkeys[$k]);
					push(@logRank, log($rank)/$logbase);
				}
			}

			# For each experiment, store the datapoints within this interval into arrays of X and Y
			push(@Xlist, join(';', @logCount));
			push(@Ylist, join(';', @logRank));
		}
	}
	return (\@Xlist, \@Ylist, \@commonIntervals);
}

1;