#---------------------------------------------------------------------------------------
# Singapore, 1st July 2017

# ACADEMIC SOFTWARE LICENSE for PLSDBC
# ************************************

# Copyright:
# Wing-Cheong Wong
# Bioinformatics Institute (BII) A*STAR Singapore

# This software is subjected to license restrictions. The software is
# provided as is. In its present form and without written consent of 
# the copyright holder, the software is provided in accordance with GPL
# (www.gnu.org/licenses/gpl.html). Among other issues, this excludes any
# warranty and indemnity when using this software. For a commercial license,
# please approach the authors (via email to wongwc@bii.a-star.edu.sg).

#---------------------------------------------------------------------------------------

#!/usr/bin/perl
package filefun;

use strict;	# Forces variables to be declared
use warnings;
use POSIX;

sub appendfile {	# Append the desired columns from file2 to file1
	my ($datafile1,$datafile2,$colidx,$outfile) = @_;

	my $i; my $j;
	my $var1; my $var2; my $var3;
	my @tmp1; my @tmp2; my @tmp3=();
	my $found; my $idx;

	my @data1; my $header1; ($var1,$header1)=readfile($datafile1); @data1=@$var1;
	my @data2; my $header2; ($var2,$header2)=readfile($datafile2); @data2=@$var2;
	my @datacol=@$colidx;

	my $newheader;
	@tmp1=split(/\t/,$header2);
	for ($i=0,@tmp2=(); $i<scalar(@datacol); $i++) { push(@tmp2,$tmp1[$datacol[$i]]); } $var2=join("\t",@tmp2);
	$newheader=$header1."\t".$var2;

	my @jointdata=();
	if (scalar(@data1)>0 && scalar(@data2)>0) {
		for ($i=0; $i<scalar(@data1); $i++) {
			@tmp1=split(/\t/,$data1[$i]);
			for ($j=0,$found=0,$idx=-1; $j<scalar(@data2); $j++) {
				@tmp2=split(/\t/,$data2[$j]);
				if (uc($tmp1[0]) eq uc($tmp2[0])) { $found=1; $idx=$j; }
			}

			if ($found==1) { @tmp2=split(/\t/,$data2[$idx]); for ($j=0,@tmp3=(); $j<scalar(@datacol); $j++) { push(@tmp3,$tmp2[$datacol[$j]]); } }
			else { for ($j=0,@tmp3=(); $j<scalar(@datacol); $j++) { push(@tmp3,0); } }
			$var3=join("\t",@tmp3);

			push(@jointdata,$data1[$i]."\t".$var3);

#			if ($i==100) { last; }	# Test code
		}
	}
#	print $newheader."\n"; for ($i=0; $i<scalar(@jointdata); $i++) { print $jointdata[$i]."\n"; }
	writefile($newheader,\@jointdata,$outfile);
}

sub selectdataColumns {
	my ($datafile,$colref) = @_;
	my @selectcol=@$colref;

	my $i; my $j;
	my $var1; my $var2;
	my @tmp1; my @tmp2;
	my @data; my $header;
	($var1,$header)=readfile($datafile); @data=@$var1;

	my $newheader;
	@tmp1=split(/\t/,$header);
	for ($i=0,@tmp2=(); $i<scalar(@selectcol); $i++) { push(@tmp2,$tmp1[$selectcol[$i]]); }
	$newheader=join("\t",@tmp2);

	my @subdata=();
	for ($i=0; $i<scalar(@data); $i++) {
		@tmp1=split(/\t/,$data[$i]);
		for ($j=0,@tmp2=(); $j<scalar(@selectcol); $j++) { push(@tmp2,$tmp1[$selectcol[$j]]); }
		push(@subdata,join("\t",@tmp2));
	}
#	print $newheader."\n"; for ($i=0; $i<scalar(@subdata); $i++) { print $subdata[$i]."\n"; }
	
	return (\@subdata,$newheader);
}

sub removedataColumn {
	my ($datafile,$colref,$outfile) = @_;
	my @excludecol=@$colref;

	my @data=();
	my ($dataref, $header)=readfile($datafile); @data=@$dataref;

	my $i; my $j; my $k;
	my @tmp1; my @tmp2;

	my @filterdata=(); my $newheader;

	if (scalar(@data)>0) {
		# Create new header
		@tmp1=split(/\t/,$header);
#		print "No.of data columns: ".scalar(@tmp1)."\n";
		for ($j=0,@tmp2=(); $j<scalar(@tmp1); $j++) {
			for ($k=0; $k<scalar(@excludecol); $k++) { if ($j!=$excludecol[$k]) { push(@tmp2,$tmp1[$j]); } }
		}
		$newheader=join("\t",@tmp2);

		# Create new data matrix
		for ($i=0; $i<scalar(@data); $i++) {
			@tmp1=split(/\t/,$data[$i]);
			for ($j=0,@tmp2=(); $j<scalar(@tmp1); $j++) {
				for ($k=0; $k<scalar(@excludecol); $k++) { if ($j!=$excludecol[$k]) { push(@tmp2,$tmp1[$j]); } }
			}
			push(@filterdata,join("\t",@tmp2));
		}

		# Write file
		writefile($newheader,\@filterdata,$outfile);
	}
}

sub addtagtoColumn {	# Append additional tags to desired column
	my ($datafile,$colidx,$tag,$outfile) = @_;

	my $i; my $j;
	my $dataRef;
	my $header;
	my @tmp1;
	my @data=();
	($dataRef,$header)=readfile($datafile); @data=@$dataRef;

	if (scalar(@data)>0) {
		for ($i=0; $i<scalar(@data); $i++) {
			@tmp1=split(/\t/,$data[$i]);
			$tmp1[$colidx]=$tag."-".$tmp1[$colidx];
			$data[$i]=join("\t",@tmp1);
		}
		writefile($header,\@data,$outfile);
	}
}

sub readfile {	# Read sequence count file and add 1 for zero entries
	my ($datafile) = @_;

	my $i;
	my $var1;
	my @tmp1;
	my $line;
	my $count=0;

	my $rowheader;
	my @data=();
	open (MYFILE1, $datafile);
	while (<MYFILE1>) {
		$line = $_;
		$line =~ s/\r//g;	# remove linefeed
		$line =~ s/\n//g;	# remove linefeed
		if ($count==0) { $rowheader=$line; }
		if ($count > 0) { push(@data, $line); }

		$count=$count+1;
	}
	close(MYFILE1);

	@tmp1=split(/\t/,$rowheader);
#	for ($i=0; $i<scalar(@tmp1); $i++) { print $tmp1[$i]."\n"; }
#	print scalar(@tmp1)."\n";
#	for ($i=0; $i<scalar(@data); $i++) {
#		@tmp1=split(/\t/,$data[$i]);
#		print scalar(@tmp1)."\n";
#	}

	return (\@data,$rowheader);
}

sub writefile {
	my ($header,$dataRef,$outfile) = @_;
	my @data=@$dataRef;

	my $i;
	my $j;
	my $var1;
	my @tmp1;

	open(MYFILE1, "> $outfile");
	print MYFILE1 $header."\n";
	for ($i=0; $i<scalar(@data); $i++) {
		print MYFILE1 $data[$i]."\n";
	}
	close(MYFILE1);
}

sub readCountfile {	# Read sequence count file and add 1 for zero entries
	my ($datafile,$numexpt) = @_;

	my $i;
	my $var1;
	my @tmp1;
	my $line;
	my $count=0;
	my $flag=0;
	my $header; my @data=();

	# filename is left empty
	if ($datafile eq "") {	return (\@data,$header); }

	open (MYFILE1, $datafile);
	while (<MYFILE1>) {
		$line = $_;
		$line =~ s/\r//g;	# remove linefeed
		$line =~ s/\n//g;	# remove linefeed
#		print $line."\n";

		if ($count==0) { $header=$line; }

		# Read the sequence count and fill the empty columns with 0 count
		if ($count > 0) {
#			print $line."\n";
			@tmp1 = split(/\t/, $line);
#			print scalar(@tmp1)."\n";
			# e.g. hsa-let-7a-3p	9	3	38	108	4	20	16		24	3		18	36	53	23	309	11	2
			for ($i=1; $i<scalar(@tmp1); $i++) {
				$tmp1[$i] =~ s/\r//g; $tmp1[$i] =~ s/\n//g;	# remove linefeed

				# Give pseudocount of 1 for missing values
				if ($tmp1[$i] eq '') { $tmp1[$i]=1; }
				else { $tmp1[$i]=$tmp1[$i]+1; }
			}
			$var1 = join("\t", @tmp1);

			# Append the length to be equal columns (based on no. of experiments)
			for ($i=scalar(@tmp1); $i<($numexpt+1); $i++) {
				$var1 = $var1."\t1";
			}
#			print $var1."\n";

			@tmp1 = split(/\t/, $var1);
			if (scalar(@tmp1) == ($numexpt+1)) { push(@data, $var1); }
			else { print "Error...in line ".$line."\n"; }
		}
		$count = $count+1;
	}
	close(MYFILE1);

	return (\@data,$header);
}

sub writeCountfile {
	my ($header,$dataRef,$outfile) = @_;
	my @data=@$dataRef;

	my $i;
	my $j;
	my $var1;
	my @tmp1;

	open(MYFILE1, "> $outfile");
	print MYFILE1 $header."\n";
	for ($i=0; $i<scalar(@data); $i++) {
		@tmp1 = split(/\t/, $data[$i]);
		print MYFILE1 $tmp1[0];
		for ($j=1; $j<scalar(@tmp1); $j++) {
			$var1=($tmp1[$j]-1); if ($var1<0) { $var1=0; }
			print MYFILE1 "\t".$var1;
#			print MYFILE1 "\t".log($tmp1[$j])/log(10);
		}
		print MYFILE1 "\n";
	}
	close(MYFILE1);
}

sub createnormCountfile {
	my ($datafile,$colidxRef,$outfile) = @_;
	my @colidx=@$colidxRef;

	my $header; my @filecontent;
	my $i;
	my $j;
	my $var1;
	my $var2;
	my @tmp1;
	my @tmp2;

	my $newheader=""; my @newfilecontent=();

	($var1,$header)=readfile($datafile); @filecontent=@$var1;

	# Create new file header
	@tmp1=split(/\t/,$header);
	for ($j=0,@tmp2=(); $j<scalar(@colidx); $j++) { push(@tmp2,$tmp1[$colidx[$j]]); }
	$newheader=join("\t",@tmp2);

	# Create new file content
	for ($i=0; $i<scalar(@filecontent); $i++) {
		@tmp1=split(/\t/,$filecontent[$i]);
		for ($j=0,@tmp2=(); $j<scalar(@colidx); $j++) { push(@tmp2,$tmp1[$colidx[$j]]); }
		push(@newfilecontent,join("\t",@tmp2));
	}

	writefile($newheader,\@newfilecontent,$outfile);
}
1;
