#---------------------------------------------------------------------------------------
# Singapore, 1st July 2017

# ACADEMIC SOFTWARE LICENSE for PLSDBC
# ************************************

# Copyright:
# Wing-Cheong Wong
# Bioinformatics Institute (BII) A*STAR Singapore

# This software is subjected to license restrictions. The software is
# provided as is. In its present form and without written consent of 
# the copyright holder, the software is provided in accordance with GPL
# (www.gnu.org/licenses/gpl.html). Among other issues, this excludes any
# warranty and indemnity when using this software. For a commercial license,
# please approach the authors (via email to wongwc@bii.a-star.edu.sg).

#---------------------------------------------------------------------------------------

use strict;	# Forces variables to be declared
use configs;
use filefun;
use mathsfun;
use biascorrAlgo;

sub ppsSolver {
	my($datafile, $numexpt, $techrepRef, $ppslistRef) = @_;
	my $i;
	my $j;
	my $k;
	my $var1;
	my $var2;
	my $var3;
	my $var4;
	my @tmp1;
	my @tmp2;

	my @origdata; my $header;
	($var1,$header)=filefun::readCountfile($datafile,$numexpt); @origdata=@$var1;	# It gives zero count a pseudovalue=1
#	for ($i=0; $i<scalar(@origdata); $i++) { print $origdata[$i]."\n"; }

	# Create a hashtable of technical replicate groups
	my @techrepgrp=@$techrepRef;
	my %hashtable1=(); my @keys1=();
	for ($i=0; $i<scalar(@techrepgrp); $i++) {
		if (exists($hashtable1{$techrepgrp[$i]})) { $hashtable1{$techrepgrp[$i]}=$hashtable1{$techrepgrp[$i]}.";".$i; }
		else { $hashtable1{$techrepgrp[$i]}=$i; }
	}
	@keys1=sort {$a<=>$b} keys %hashtable1;
	for ($i=0; $i<scalar(@keys1); $i++) { print "TechRep set[".$i."]:".$hashtable1{$keys1[$i]}."[columns]\n"; } print "\n";

	my @mincountPerExpt; my @maxcountPerExpt; my @totalreadsPerExpt; my @maxcountPerExptIdx;
	($var1,$var2,$var3,$var4)=biascorrUtil::createExptDataSummary(\@origdata,$numexpt,0);	# Find the smallest, largest read count and total reads for each experiment
	@mincountPerExpt=@$var1; @maxcountPerExpt=@$var2; @totalreadsPerExpt=@$var3; @maxcountPerExptIdx=@$var4;

	my @techrepsetRef=();
	for ($i=0; $i<scalar(@keys1); $i++) {
		@tmp1=split(/\;/,$hashtable1{$keys1[$i]});
		for ($j=0,@tmp2=(); $j<scalar(@tmp1); $j++) { push(@tmp2,$totalreadsPerExpt[$tmp1[$j]]); } @tmp2=sort {$b<=>$a} @tmp2;		# Find the largest count within this technical rep set
		for ($j=0; $j<scalar(@tmp1); $j++) { if ($totalreadsPerExpt[$tmp1[$j]]==$tmp2[0]) { push(@techrepsetRef,$tmp1[$j]); last; } }	# Store the column of the largest read count within this technical rep set
	}
	print "Columns with largest total reads:\n"; for ($i=0; $i<scalar(@techrepsetRef); $i++) { print "[TechRep set[".$i."]:".$techrepsetRef[$i]."\t"; } print "\n\n";

	my @pps=@$ppslistRef;	# The set of points-per-segment to test for best technical replicate correlation
#	my @pps=(3,4,5,10,15,20,25,30,35,40,45,50,55,60,65,70,80,90,100);
#	my @pps=(50,100,150,200,250,300);	# For mRNA
	my @avgRsq=();

	my $logbase=log($configs::BASEOFLOG);
	my @intervallist=(); my $maxinterval=-1;
	my @numptspersegment=(); my @intervalsperExpt=(); my @regressionlist=();
	my @data=(); my @subdata=(); my @logFCsubdata=();
	my @xarray=(); my @yarray=();
	my @refcountPerExpt=();
	my $slope; my $intercept; my $Rsq; my $stdslope;
	my @refSlope=();

	for ($k=0; $k<scalar(@pps); $k++) {
		print "Computing PPS=".$pps[$k]."...\n";

		# Reset data back to the original for every loop
		@data=(); push(@data,@origdata);

		# To generate the maxcountperexpt parameter for function biascorrAlgo::scaleDatatoZipLaw
		($var1,$var2,$var3,$var4)=biascorrUtil::createExptDataSummary(\@data,$numexpt,1);	# Find the smallest, largest read count and total reads for each experiment
		@mincountPerExpt=@$var1; @maxcountPerExpt=@$var2; @totalreadsPerExpt=@$var3; @maxcountPerExptIdx=@$var4;

		for ($i=0,@numptspersegment=(); $i<$numexpt; $i++) { push(@numptspersegment,$pps[$k]); }
		($var1,$var2) = biascorrUtil::computeIntervalsByPtsperSeg(\@data, $numexpt, $logbase, \@numptspersegment); @intervallist=@$var1; $maxinterval=$var2;

		# Get the regression parameters of the first segment per experiment of the log-log plot
		for ($i=0,@intervalsperExpt=(); $i<scalar(@intervallist); $i++) {
			@tmp1=split(/\;/,$intervallist[$i]);
			if ($tmp1[1]==0) { push(@intervalsperExpt,$intervallist[$i]); }
		}
		($var1,$var2)=biascorrAlgo::computeRegressionParameters(\@data, $numexpt, \@intervalsperExpt, $logbase);
		@regressionlist=@$var1; @intervalsperExpt=@$var2;
		print "Regression list for segment[0] per expt[numexpt;segmentID;range;slope;intercept;Rsq;size]:\n";
		for ($i=0; $i<scalar(@regressionlist); $i++) { print $regressionlist[$i]."\n"; } print "\n";

		# Compute the median slope of segment[0] among all experiments
		for ($i=0,@tmp2=(); $i<scalar(@regressionlist); $i++) { @tmp1=split(/\;/,$regressionlist[$i]); push(@tmp2,$tmp1[3]); }
		push(@refSlope,mathsfun::getmedian(\@tmp2));


		for ($i=0; $i<$numexpt; $i++) {
			# Create a sublist of interval for per experiment's log-log plot. Process each segmented log-log plot a time
			for ($j=0,@intervalsperExpt=(); $j<scalar(@intervallist); $j++) {
				@tmp1 = split(/\;/, $intervallist[$j]);	# intervallist ~ [expt;segmentid;startvalue;stopvalue]
				if ($tmp1[0]==$i) { push(@intervalsperExpt, $intervallist[$j]); }
			}

			if (scalar(@intervalsperExpt)>0) {	# For each segment of the same experiment
				# Obtain regression parameters for each interval
				($var1,$var2)=biascorrAlgo::computeRegressionParameters(\@data, $numexpt, \@intervalsperExpt, $logbase);
				@regressionlist=@$var1; @intervalsperExpt=@$var2;

				# Perform scaling of datapoints per experiment
				if (scalar(@regressionlist)>0) {
					print "Regression list for expt[".$i."] per segment[numexpt;segmentID;range;slope;intercept;Rsq;size]:\n";
					for ($j=0; $j<scalar(@regressionlist); $j++) { print $regressionlist[$j]."\n"; }
					$var1=biascorrAlgo::scaleDatatoZipfLaw(\@data, \@regressionlist, $maxcountPerExpt[$i], -1, $configs::BASEOFLOG);
					@data=@$var1;
				}
			}
		}

		# Remove pseudocount of 1
		@subdata=(); $var1=biascorrUtil::filterZeroCountData(\@data); @subdata=@$var1;
		print "\nDatasize after removing zero counts:".scalar(@subdata)."\n\n";

		# Find the smallest, largest read count and total reads for each experiment
		@mincountPerExpt=@maxcountPerExpt=@totalreadsPerExpt=@maxcountPerExptIdx=();
		($var1,$var2,$var3,$var4)=biascorrUtil::createExptDataSummary(\@subdata,$numexpt,1);
		@mincountPerExpt=@$var1; @maxcountPerExpt=@$var2; @totalreadsPerExpt=@$var3; @maxcountPerExptIdx=@$var4;

		# Get the reference count for each experiment based on the reference array of each technical replicate set
		for ($i=0,@refcountPerExpt=(); $i<scalar(@keys1); $i++) {
			@tmp1=split(/\;/,$hashtable1{$keys1[$i]});
			for ($j=0; $j<scalar(@tmp1); $j++) {
				$var1=biascorrUtil::selectDataCol(\@subdata,$tmp1[$j]+1); @xarray=@$var1;
				push(@refcountPerExpt,$xarray[$maxcountPerExptIdx[$tmp1[0]]]);
			}
		}
#		for ($i=0; $i<scalar(@refcountPerExpt); $i++) { print "Expt[".$i."]\t".$refcountPerExpt[$i]."\n"; }

		# Convert data to log2 FC
		for ($i=0,@logFCsubdata=(); $i<scalar(@subdata); $i++) {
			@tmp1=split(/\t/, $subdata[$i]);
			# offset by 1 due to gene transcript name column
			for ($j=1; $j<scalar(@tmp1); $j++) {
				$tmp1[$j]=log($tmp1[$j]/$refcountPerExpt[$j-1])/log(2);
				$tmp1[$j]=sprintf("%.5f",$tmp1[$j]);
			}
			$var1=join("\t",@tmp1);
			push(@logFCsubdata,$var1);
		}
#		for ($i=0; $i<scalar(@logFCsubdata); $i++) { print $i."\t".$logFCsubdata[$i]."\n"; }

#		@keys1 = keys %hashtable1;
#		for ($i=0,@tmp2=(); $i<scalar(@keys1); $i++) {
#			@tmp1=split(/\;/,$hashtable1{$keys1[$i]});
#			$var1=biascorrUtil::selectDataCol(\@logFCsubdata,$tmp1[0]+1); @xarray=@$var1;
#			print "Comparison\tSlope\tIntercept\tRsquare\tSTD[slope]\n";
#			for ($j=1; $j<scalar(@tmp1); $j++) {
#				$var2=biascorrUtil::selectDataCol(\@logFCsubdata,$tmp1[$j]+1); @yarray=@$var2;
#				($slope,$intercept,$Rsq,$stdslope)=mathsfun::linearRegression(\@xarray,\@yarray,'');
#				print $tmp1[0]."vs".$tmp1[$j]."\t".$slope."\t".$intercept."\t".$Rsq."\t".$stdslope."\n";
#				push(@tmp2,$Rsq);
#			}
#		}

		@keys1 = keys %hashtable1;
		for ($i=0,@tmp2=(); $i<scalar(@keys1); $i++) {
			@tmp1=split(/\;/,$hashtable1{$keys1[$i]});
			$var1=biascorrUtil::selectDataCol(\@logFCsubdata,$techrepsetRef[$i]+1); @xarray=@$var1;
			print "Comparison\tSlope\tIntercept\tRsquare\tSTD[slope]\n";
			for ($j=0; $j<scalar(@tmp1); $j++) {
				if ($tmp1[$j]!=$techrepsetRef[$i]) {
					$var2=biascorrUtil::selectDataCol(\@logFCsubdata,$tmp1[$j]+1); @yarray=@$var2;
					($slope,$intercept,$Rsq,$stdslope)=mathsfun::linearRegression(\@xarray,\@yarray,'');
					print $techrepsetRef[$i]."vs".$tmp1[$j]."\t".$slope."\t".$intercept."\t".$Rsq."\t".$stdslope."\n";
					push(@tmp2,$Rsq);
				}
			}
		}
		($var1,$var2)=mathsfun::getSummaryStatistics(\@tmp2);	# mean and stdev
		push(@avgRsq,$var1);
		print "\n";
	}
	print "PPS\tavgRsquare\trefSlope[median]:\n";
	for ($k=0; $k<scalar(@pps); $k++) { print $pps[$k]."\t".$avgRsq[$k]."\t".$refSlope[$k]."\n"; } print "\n";

	# Find the pps setting with the highest average correlation across the technical replicate sets
	my $maxRsq=0; my $optpps=0; my $optrefSlope=-1;
	for ($k=0; $k<scalar(@pps); $k++) {
		if ($avgRsq[$k]>$maxRsq) { $maxRsq=$avgRsq[$k]; $optpps=$pps[$k]; $optrefSlope=$refSlope[$k]; }
	}
	$optrefSlope=sprintf("%.2f",$optrefSlope);

	print "[Recommended setting] PPS:".$optpps.",RefSlope:".$optrefSlope."\n";

	return ($optpps,$optrefSlope);
}
1;