#---------------------------------------------------------------------------------------
# Singapore, 1st July 2017

# ACADEMIC SOFTWARE LICENSE for PLSDBC
# ************************************

# Copyright:
# Wing-Cheong Wong
# Bioinformatics Institute (BII) A*STAR Singapore

# This software is subjected to license restrictions. The software is
# provided as is. In its present form and without written consent of 
# the copyright holder, the software is provided in accordance with GPL
# (www.gnu.org/licenses/gpl.html). Among other issues, this excludes any
# warranty and indemnity when using this software. For a commercial license,
# please approach the authors (via email to wongwc@bii.a-star.edu.sg).

#---------------------------------------------------------------------------------------

#!/usr/bin/perl
use strict;	# Forces variables to be declared
use warnings;
use configs;

use filefun;
use filterDataCountByGrp;
use ppsSolver;
use alignSeqCountData;
use alignSeqCountDatawithOutliers;

my $i; my $j;

# ---------------------------------------------------------------------------------------------------------- #
# Dilution data
# ---------------------------------------------------------------------------------------------------------- #
# NUGC30078-12p;NUGC30078-6p;NUGC30078-3p;NUGC30078-1.5p;NUGC30081-12p;NUGC30081-6p;NUGC30081-3p;NUGC30081-1.5p;NUGC30082-5p
# AGS0078-12p;AGS0078-3p;AGS0081-12p;AGS0081-3p;AGS0082-5p

# For NUGC3 only analysis
#my @ppslist=(3,4,5,10,15,20,25,30,35,40,45,50,55,60,65,70,80,90,100);
#my @techrepgroup=(1,1,1,1,1,1,1,1);	# zero refers to non-processing columns
#my ($pps,$refslope)=ppsSolver($configs::DILUTIONDATA.'dilutionrun007881bowtie1NUGC3.txt',8,\@techrepgroup,\@ppslist);
#alignSeqCountData($configs::DILUTIONDATA.'dilutionrun007881bowtie1NUGC3.txt',8,45,-0.88);

# For NUGC3+AGS analysis
# Perform filtering on data set
#my @techrepgroup=(1,1,1,1,1,1,1,1,2,2,2,2); my @exptgroup=();	# Each cellline as a group
#filterDataCountByGrp($configs::DILUTIONDATA.'dilutionrun007881bowtie1-all.txt',12,\@techrepgroup,\@exptgroup,$configs::DILUTIONDATA.'dilutionrun007881bowtie1.txt');
#filterDataCountByGrp($configs::DILUTIONDATA.'dilutionrun007881bowtie2global-all.txt',12,\@techrepgroup,\@exptgroup,$configs::DILUTIONDATA.'dilutionrun007881bowtie2global.txt');
#filterDataCountByGrp($configs::DILUTIONDATA.'dilutionrun007881bwa-all.txt',12,\@techrepgroup,\@exptgroup,$configs::DILUTIONDATA.'dilutionrun007881bwa.txt');
#filterDataCountByGrp($configs::DILUTIONDATA.'dilutionrun007881novo-all.txt',12,\@techrepgroup,\@exptgroup,$configs::DILUTIONDATA.'dilutionrun007881novo.txt');

# To determine the best PPS setting for Zipf-correction
#my @ppslist=(3,4,5,10,15,20,25,30,35,40,45,50,55,60,65,70,80,90,100);
#my @techrepgroup=(1,1,1,1,1,1,1,1,2,2,2,2);	# zero refers to non-processing columns
#my ($pps,$refslope)=ppsSolver($configs::DILUTIONDATA.'dilutionrun007881bowtie1.txt',12,\@techrepgroup,\@ppslist);
#my ($pps,$refslope)=ppsSolver($configs::DILUTIONDATA.'dilutionrun007881bowtie2global.txt',12,\@techrepgroup,\@ppslist);
#my ($pps,$refslope)=ppsSolver($configs::DILUTIONDATA.'dilutionrun007881bwa.txt',12,\@techrepgroup,\@ppslist);
#my ($pps,$refslope)=ppsSolver($configs::DILUTIONDATA.'dilutionrun007881novo.txt',12,\@techrepgroup,\@ppslist);

#alignSeqCountData($configs::DILUTIONDATA.'dilutionrun007881bowtie1.txt',12,55,-0.83);
#alignSeqCountData($configs::DILUTIONDATA.'dilutionrun007881bowtie2global.txt',12,55,-0.83);
#alignSeqCountData($configs::DILUTIONDATA.'dilutionrun007881novo.txt',12,55,-0.83);
#alignSeqCountData($configs::DILUTIONDATA.'dilutionrun007881bwa.txt',12,55,-0.83);


# ---------------------------------------------------------------------------------------------------------- #
# Spike-in data (Briefing in Bioinformatics 2015 doi:10.1093/bib/bbv019)
# ---------------------------------------------------------------------------------------------------------- #
# Analysis : Filter->PL-corrected->normalization->ROCplot

# Perform filtering on data set
#my @techrepgroup=(1,2,3,4,5,6,7,8,9,10,11,12); my @exptgroup=();	# Each sample as a condition
#filterDataCountByGrp($configs::SPIKEINDATA.'GSE67074baseall.txt',12,\@techrepgroup,\@exptgroup,$configs::SPIKEINDATA.'GSE67074base.txt');	# BIB2015

# To determine the best PPS setting for Zipf-correction
#my @techrepgroup=(1,1,1,1,1,1,1,1,1,1,1,1);	# zero refers to non-processing columns
#my @ppslist=(3,4,5,10,15,20,25,30,35,40,45,50,55,60,65,70,80,90,100);
#my ($pps,$refslope)=ppsSolver($configs::SPIKEINDATA.'GSE67074base.txt',12,\@techrepgroup,\@ppslist);	# [Recommended setting] PPS:10,RefSlope:-0.95

# PL-correction
#alignSeqCountDatawithOutliers($configs::SPIKEINDATA.'GSE67074base.txt',$configs::SPIKEINDATA.'GSE67074spikein.txt',12,10,-0.95);
#alignSeqCountDatawithOutliers($configs::SPIKEINDATA.'GSE67074base.txt',$configs::SPIKEINDATA.'GSE67074spikein.txt',12,20,-0.94);

